import numpy as np
import scipy as sp

import matplotlib as mpl
from matplotlib.figure import Figure
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import *

from DNMR.tab import *

class TabFourierTransform(Tab):
    def __init__(self, data_widgets, parent=None):
        super(TabFourierTransform, self).__init__(data_widgets, 'tab_ft', parent)
        
        self.data = (np.array([]), np.array([]))
        self.left_pivot = 1
        self.right_pivot = -1
        self.moving_left = True

    def generate_layout(self):
        self.canvas.mpl_connect('button_press_event', self.process_button)

    def process_button(self, event):
        if(event.button == 1):
            if not(event.xdata is None):
                if(self.moving_left):
                    self.left_pivot = event.xdata
                else:
                    self.right_pivot = event.xdata
                self.moving_left = not(self.moving_left)
            self.update()

    def plot_logic(self):
        index = self.fileselector.spinbox_index.value()

        times = self.data_widgets['tab_phase'].data[0]
        complexes = self.data_widgets['tab_phase'].data[1]

        timespacing = (times[index][1]-times[index][0])
    
        time_index = np.argmin(np.abs(self.fileselector.data['peak_locations'][:,None] - times), axis=1)
        s_complexes = np.zeros_like(complexes)
        s_times = np.zeros_like(times)
        for i in range(s_complexes.shape[0]):
            s_complexes[i,:] = np.roll(complexes[i,:], -time_index[i])
            s_times[i,:]     = np.roll(times[i,:],     -time_index[i])
        s_reals = np.real(s_complexes)
        s_imags = np.imag(s_complexes)

        fftfreq = np.fft.fftshift(np.fft.fftfreq(s_complexes[index].shape[0], d=timespacing)) # microseconds
        fft = np.zeros_like(s_complexes)
        for i in range(s_complexes.shape[0]):
            fft[i] = np.fft.fftshift(np.fft.fft(s_complexes[i]))

        self.data = (fftfreq, fft)
        
        self.ax.plot(fftfreq, np.real(fft[index]), 'r', alpha=0.6, label='R')
        self.ax.plot(fftfreq, np.imag(fft[index]), 'b', alpha=0.6, label='I')
        self.ax.plot(fftfreq, np.abs(fft[index]), 'k', alpha=0.3, label='abs')
        
        self.ax.set_xlabel('frequency (MHz)')

        self.ax.axvline(self.left_pivot, color='k')
        self.ax.axvline(self.right_pivot, color='k')

    def get_exported_data(self):
        index = self.fileselector.spinbox_index.value()
        return { 'times': self.data_widgets['tab_phase'].data[0][index],
                 'complexes': self.data_widgets['tab_phase'].data[1][index],
                 'frequencies (MHz)': self.data[0],
                 'fft': self.data[1][index],
               }


import numpy as np
import h5py as hdf
import pytnt as tnt
import traceback
import re

from DNMR.fileops_loaders.data_struct import data_struct, hdf_to_dict

def read_hdf_valpha(file):
    toplevel = file.keys()
    points = []
    point_indices = []
    point_numbers = []
    print(toplevel)
    count = 0
    for i in toplevel: # get all points
        m = re.match('(point|entry)(?P<index>[0-9]+)', i)
        if not(m is None):
            points += [ m[0] ]
            point_numbers += [ int(m['index']) ]
            point_indices += [ count ]
            count += 1
    points = np.array(points)
    point_indices = np.array(point_indices)
    point_numbers = np.array(point_numbers)
    
    sorted_indices = np.argsort(point_numbers)
    points = points[sorted_indices]
    point_numbers = point_numbers[sorted_indices]
    
    data = data_struct()
    # load the first one, to get sizes etc.
    for key, val in file[points[0]].items():
        if(key[:5] == 'tnmr_'):
            key = key[5:]
        data[key] = [ None ] * len(points)
        
    data['size'] = len(points)

    for i, index in zip(points, point_indices):
        for key, val in file[i].items():
            if(key[:5] == 'tnmr_'):
                key = key[5:]
            # legacy logic
            if(key == 'relaxation_time'):
                key = 'delay_time'
            # end of legacy logic
            data[key][index] = val
    
    for key, val in data.items():
        # check if we can turn it into a dict, then numpy array
        try:
            ds = data_struct(hdf_to_dict(val[0]))
            for i in range(1, len(val)):
                ds += data_struct(hdf_to_dict(val[i]))
            data[key] = ds
        except:
            try:
                arr = np.array(val)
                try:
                    arr_f = arr.astype(float)
                    arr_i = arr.astype(int)
                    if(arr_i == arr_f):
                        arr = arr_i
                    else:
                        arr = arr_f
                except:
                    pass
                data[key] = arr
            except:
                pass
    if(data['times'].shape != data['reals'].shape):
        l = data['reals'].shape[1]
        if(data['params']['acquisition_time'].ndim > 0):
            data['times'] = data['params']['acquisition_time'][:,None] * np.array([ i/l for i in range(0, l) ])[None,:] # really legacy
        else:
            data['times'] = data['params']['acquisition_time'] * np.array([ i/l for i in range(0, l) ])[None,:] # really legacy

    return data



import numpy as np
import scipy as sp

import matplotlib as mpl
from matplotlib.figure import Figure
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import *

import DNMR.fileops as fileops
from DNMR.miniwidgets import *

from DNMR.tab import Tab

class TabT1Fit(Tab):
    output_frames = {}

    def __init__(self, data_widgets, parent=None):
        super(TabT1Fit, self).__init__(data_widgets, 'tab_t1_fitting', parent)
        
        self.data = (np.array([]), np.array([]))
        self.plot_data = (np.array([]), np.array([]))
        self.x0 = None
        self.sigmas = None

    def generate_layout(self):
        self.combobox_fittingroutine = QComboBox()
        self.combobox_fittingroutine.addItem('7/2 Spin')
        
        self.pushbutton_fit = QPushButton('Fit')
        self.pushbutton_fit.clicked.connect(self.fit)
        
        self.checkbox_normalize = QCheckBox('Normalize?')

        l = QHBoxLayout()
        lv = QVBoxLayout()
        lv.addWidget(self.combobox_fittingroutine)
        lv.addWidget(self.checkbox_normalize)
        l.addLayout(lv)
        l.addWidget(self.pushbutton_fit)

        def add_fit_frame(name, *args):
            # fit output
            frm = QFrame() # TODO: Make this better.
            frm.hide()
            lo = QVBoxLayout()
            self.output_frames[name] = [ frm ]
            for i in range(len(args)//2):
                w = QLabel('fitting...')
                fix = QCheckBox('Fix?')
                li = QHBoxLayout()
                li.addWidget(fix)
                li.addWidget(w)
                lo.addLayout(li)
                self.output_frames[name] += [ {'widget': w, 'label': args[2*i], 'units': args[2*i+1], 'fixed_widget': fix } ]
                
            frm.setLayout(lo)
            l.addWidget(frm)

        # Title, var_name, var_units, var_name, var_units, ...
        add_fit_frame('7/2 Spin', '\u03b30', '', 's', '', 'T1', '\u03bcs', 'r', '')

        return l

    def plot_logic(self):
        freq = self.data_widgets['tab_ft'].data[0]
        ft   = self.data_widgets['tab_ft'].data[1]
        real = np.real(ft)
        try:
            del_times = self.fileselector.data.sequence['0'].delay_time
        except:
            del_times = self.fileselector.data.sequence['0'].relaxation_time # Legacy
            print(del_times)

        integrations = np.zeros(real.shape[0])
        start_index = np.argmin(np.abs(self.data_widgets['tab_ft'].left_pivot - freq))
        end_index = np.argmin(np.abs(self.data_widgets['tab_ft'].right_pivot - freq))
        if(end_index < start_index):
            tmp = start_index
            start_index = end_index
            end_index = tmp

        integrations = np.sum(real[:,start_index:end_index], axis=1)
        if(self.checkbox_normalize.isChecked()):
            integrations /= np.max(integrations)
        rt = np.real(self.data_widgets['tab_ft'].data[1])
        uncertainties = np.ones_like(integrations)*1e-3#np.std(rt, axis=1)/np.max(np.abs(rt), axis=1)*integrations # TODO: Figure out real stddevs
        uncertainties = np.abs(uncertainties)
            
        sort_indices = np.argsort(del_times)
        del_times = del_times[sort_indices]
        integrations = integrations[sort_indices]
        uncertainties = uncertainties[sort_indices]
        
        self.ax.set_xscale('log')
        self.ax.set_xlabel('delay time (us)')
        self.ax.errorbar(del_times, integrations, label='integrations', linestyle='', marker='o', yerr=uncertainties)
        
        post_aq_max = np.max(self.fileselector.data.params.post_acquisition_time * 1e3) # this is in ms. Our axes in us
        self.ax.axvline(post_aq_max, linestyle='--', color='k')

        self.data = (del_times, integrations, uncertainties)

        if(self.plot_data[0].shape[0] > 0):
            params_list = ''
            out_frame = self.output_frames[self.combobox_fittingroutine.currentText()]
            for i in out_frame[1:]:
                params_list += f'{i['widget'].text()}\n'
            params_list = params_list[:-1]
            self.ax.plot(self.plot_data[0], self.plot_data[1], label=params_list)
        
    def fit(self):
        bounds = None
        for key, val in self.output_frames.items():
            val[0].hide()
        out_frame = self.output_frames[self.combobox_fittingroutine.currentText()]
        out_frame[0].show()
        try:
            del_times = self.fileselector.data.sequence['0'].delay_time
        except:
            del_times = self.fileselector.data.sequence['0'].relaxation_time # Legacy, as I didn't know what this was when I wrote it. Surprise, surprise

        if(self.combobox_fittingroutine.currentText() == '7/2 Spin'):
            # DEVELOPER NOTE: If you want to add more options for this, make sure to define fit_func (similarly to below) and add an item in the generate_layout function
            
            bounds = [ [0, np.max(self.data[1])*10], [-1, 10], [np.min(del_times)/10, np.max(del_times)*10], [0.99*0, 1.01*10] ]
            def fit_func(args, x):
                gamma_0 = args[0]
                s = args[1] # inversion
                T1 = args[2] # relaxation time (actual fit variable, really)
                r = args[3] # stretched exponent (ideally 1)
                #y = y0 (1-(1+s) ((1/84)*Exp[-(t/T1)^r]+(3/44)*Exp[-(6 t/T1)^r]+(75/364)*Exp[-(15 t/T1)^r]+(1225/1716)*Exp[-(28 t/T1)^r]))
                fit = gamma_0 * (1-(1+s)*(
                                            (1/84)*     np.exp(-np.pow(x/T1,    r)) + 
                                            (3/44)*     np.exp(-np.pow(6*x/T1,  r)) +
                                            (75/364)*   np.exp(-np.pow(15*x/T1, r)) +
                                            (1225/1716)*np.exp(-np.pow(28*x/T1, r)) 
                                         ))
                return fit
        def cost_func(args, x, y, yerr):
            return np.sum(np.square((fit_func(args, x) - y)/np.maximum(yerr, 0.01))) # more points is more fits

        if(self.checkbox_normalize.isChecked()):
            if(np.abs(np.max(self.data[1]) - 1.0) > 1e-3):
                d0 = self.data[0]
                d1 = self.data[1] / np.max(self.data[1])
                d2 = self.data[2] / np.max(self.data[1])
                self.data = (d0, d1, d2)
            
        if(self.x0 is not None):
            for i in range(len(self.x0)):
                for bi in range(len(bounds)):
                    if(out_frame[i+1]['fixed_widget'].isChecked()):
                        # Fix
                        bounds[bi] = sp.optimize.Bounds(self.x0[bi], self.x0[bi])
            
        # global minimum
        res = sp.optimize.differential_evolution(lambda x: cost_func(x, 
                                                                     self.data[0], 
                                                                     self.data[1],
                                                                     self.data[2]), 
                                                 bounds=bounds)
        # get uncertainties on the fit, as I am too lazy to do the full analysis when scipy will do it for me
        try:
            popt, pcov = sp.optimize.curve_fit(lambda xs, *args: fit_func(args, xs), self.data[0], self.data[1], p0=res.x)
        
            print(res)
            self.x0 = popt
            self.sigmas = np.sqrt(np.diag(pcov))
            x_vals = self.data[0]
            if(x_vals.shape[0] < 100):
                x_vals = np.exp(np.linspace(np.log(np.min(x_vals)), np.log(np.max(x_vals)), 100, endpoint=True))
            self.plot_data = (x_vals, fit_func(popt, x_vals))
            for i in range(len(self.x0)):
                try:
                    digits = int(np.ceil(np.abs(np.log10(self.sigmas[i]))))
                except:
                    digits = 10000 # sigma negative - a sign that something has gone horribly wrong and the user should deal with the drama. Show them the digits.
                if(self.sigmas[i] > 1.0):
                    rounded_digits = -digits+1
                else:
                    rounded_digits = digits
                    
                display_sigma = np.round(self.sigmas[i], rounded_digits)
                display_x = np.round(self.x0[i], rounded_digits)
                
                out_frame[i+1]['widget'].setText(f'{out_frame[i+1]["label"]}={display_x:,}\u00b1{display_sigma:,}{out_frame[i+1]["units"]}')
        except Exception as e:
            traceback.print_exc()
        self.update()
        
    def get_exported_data(self):
        out_frame = self.output_frames[self.combobox_fittingroutine.currentText()][1:]
        params_dict = {}
        if(self.x0 is not None):
            cnt = 0
            for i in out_frame:
                params_dict[i['label']] = [ str(self.x0[cnt]) + ' ' + i['units'] ]
                params_dict[i['label']+' error'] = [ str(self.sigmas[cnt]) + ' ' + i['units'] ]
                cnt += 1
        
        index = self.fileselector.spinbox_index.value()
        pd = {
                 'frequencies (MHz)': self.data_widgets['tab_ft'].data[0],
                 'fft': self.data_widgets['tab_ft'].data[1][index],
                 'delays': self.data[0],
                 'integrals': self.data[1],
                }
        pd.update(params_dict)
        return pd



import numpy as np
import scipy as sp

import matplotlib as mpl
from matplotlib.figure import Figure
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import *

from DNMR.miniwidgets import *
from DNMR.tab import Tab

class TabFieldScan(Tab):
    def __init__(self, data_widgets, parent=None):
        super(TabFieldScan, self).__init__(data_widgets, 'tab_fieldscan', parent)
        
        self.data = ([], [])

    def plot_logic(self):
        if('ppms_mf' in self.fileselector.data.keys() or 'ppms_field' in self.fileselector.data.keys()):
            index = self.fileselector.spinbox_index.value()

            times = self.data_widgets['tab_phase'].data[0]
            complexes = self.data_widgets['tab_phase'].data[1]

            reals = np.real(complexes)
            imags = np.imag(complexes)
            
            real_integral = np.sum(reals, axis=1)
            imag_integral = np.sum(imags, axis=1)
            mag_integral = np.sum(np.abs(reals + 1j*imags), axis=1)

            try:
                fields = self.fileselector.data.ppms_field
            except:
                try:
                    fields = self.fileselector.data.ppms_mf
                except:
                    return
            self.ax.plot(fields, np.abs(real_integral + 1j*imag_integral), 'k', alpha=0.6, label=f'Mag. \u222b', linestyle='None', marker='o')
            #self.ax.plot(fields, real_integral, 'r', alpha=0.6, label='R \u222b', linestyle='None', marker='o')
            #self.ax.plot(fields, imag_integral, 'b', alpha=0.6, label='I \u222b', linestyle='None', marker='o')
            
            self.ax.set_xlabel('field (T)')
            
            pvt = self.fileselector.data['peak_locations'][index]
            max_index = np.argmin(np.abs(times[0] - pvt))
            r = reals[:,max_index]
            i = imags[:,max_index]
            M = np.sqrt(np.square(r) + np.square(i))
            self.ax.plot(fields, M, linestyle='None', marker='x', color='k', label='Mag. (peak)')
            #self.ax.plot(fields, r, linestyle='None', marker='x', color='r', label='R (peak)')
            #self.ax.plot(fields, i, linestyle='None', marker='x', color='b', label='I (peak)')
            
            self.data[0] = fields
            self.data[1] = M


    def get_exported_data(self):
        index = self.fileselector.spinbox_index.value()
        return { 'times': self.data_widgets['tab_phase'].data[0][index],
                 'complexes': self.data_widgets['tab_phase'].data[1][index],
                 'fields': self.data[0],
                 'magnitudes': self.data[1],
               }